"""
Settings for the JSL-Notebook-Preprocessor

These settings are applied for Pre-Processing on Entire Notebook Cells.
One Notebook consists of multiple Cells and a Cell consists of Multiple Lines seperated by \n
There are 2 Levels of Settings :
Cell Based Rules : Treats entire Cell as a string and applies Rule
Line Based Settings : Applies rule to lines of code in cell individually.

Note:
Lines are derived by splitting on \n are not always valid python code, since actual code can be multi-line
"""

# Print content of cells before running them or not

from johnsnowlabs.utils.testing.nb_nodes import Nodes

print_cell_before_executing = False

inject_header_nodes = [Nodes.imports]

# Applied to entire Cell. Handy when Code appears in Multi-Line Formats (Like Pipeline Definitions)
# Only first Match is applied if multiple keys are matches
sub_string_replacement_cells = {
    # 'spark = jsl.start()': 'spark = nlp.start()',
    # This rule means Replace 'Pipeline(' with 'nlp.Pipeline('
    # 'Pipeline(': 'nlp.Pipeline(',
}

sub_string_replacement_lines = {
    # This rule means if a single line contains '/content/' replace '/content/' with './'
    "/content/": "./",
    # 'spark = jsl.start()': 'spark = nlp.start()',
    # 'spark = jsl.start()': 'spark = nlp.start()',
}

# Hard matches Line Drops
equal_string_drop_lines = [
    # This means if a cell contains a line with 'jsl.install()', drop that line but keep rest of cell
    "nlp.install()",
]

# Hard matches replace
equal_string_replacement_lines = {
    # This means if a cell contains a line equal to 'spark = jsl.start()' replace it with 'spark = nlp.start();from nlp import *'
    # 'spark = nlp.start()': 'spark = nlp.start()',
    "spark = nlp.start()": 'spark = nlp.start(model_cache_folder="/home/ckl/dump/cache_pretrained")',
    # CONVERSION FIXES
}

sub_string_drop_lines = [
    # This means if a line in a cell contains 'files.upload()' drop that line but keep the rest
    "files.upload()",
    # 'get_ipython',
    "pip install",
    "pip -q",
    "from google",
    "google.colab",
    "google.",
    "colab",
    "nlp.install",
    "license_keys",
    "spark_ocr.json",
    "spark_jsl.json",
    "plt.",
]

# These Regexes are applied to the entire cell as very last step, after applying all rules
regex_cell_content_drop_matcher = [
    # This rule means drop any lines matching r'spark = sparkocr.start\(.*?\)' but keep rest of cell intact
    r"spark = sparkocr.start\(.*?\)",
    r"spark = sparknlp_jsl.start\(.*?\)",
]

#### General Testing Configs and Folders
# testing_dir = f'{settings.root_dir}/tmp_tests'
success_worker_print = "$$JSL_TESTING_WORKER_SUC$$"
workshop_git = "https://github.com/JohnSnowLabs/spark-nlp-workshop.git"
testing_dir = (
    "/home/ckl/Documents/freelance/jsl/johnsnowlabs-4-real/tests/test_notebooks/workdir"
)
tmp_notebook_dir = f"{testing_dir}/notebook_tests"
tmp_py_script_dir = f"{testing_dir}/notebook_tests"
tmp_markdown_dir = f"{testing_dir}/markdown_tests"
workshop_local_folder = f"{tmp_notebook_dir}/spark-spark_nlp-workshop"
workshop_cert_nb_folder = (
    f"{workshop_local_folder}/tutorials/Certification_Trainings_JSL"
)
workshop_fin_folder = f"{workshop_cert_nb_folder}/Finance"
workshop_leg_folder = f"{workshop_cert_nb_folder}/Healthcare"
workshop_med_folder = f"{workshop_cert_nb_folder}/Legal"
workshop_pub_folder = f"{workshop_cert_nb_folder}/Public"
testing_work_dir = f"{tmp_notebook_dir}/executed_notebooks"



from johnsnowlabs import *


def run_test():
    spark = nlp.start()
    doc = nlp.DocumentAssembler().setInputCol("text").setOutputCol("doc")

    tok = nlp.Tokenizer().setInputCols("doc").setOutputCol("tok")

    embeddings = (
        nlp.WordEmbeddingsModel.pretrained("glove_100d", "en")
        .setInputCols("doc", "tok")
        .setOutputCol("embeddings")
    )

    ner = (
        nlp.NerDLModel.pretrained("nerdl_fewnerd_100d")
        .setInputCols(["doc", "tok", "embeddings"])
        .setOutputCol("ner")
    )

    ner_converter = (
        nlp.NerConverter().setInputCols(["doc", "tok", "ner"]).setOutputCol("ner_chunk")
    )

    text = "Peter Parker is a nice guy and lives in New York"
    spark_df = spark.createDataFrame([[text]]).toDF("text")

    p = Pipeline(stages=[doc, tok, embeddings, ner, ner_converter])
    p.fit(spark_df).transform(spark_df).show()



import re

from nbclient import NotebookClient
from nbclient.exceptions import CellExecutionError  # noqa
from nbconvert.preprocessors import ExecutePreprocessor
from nbformat import NotebookNode

from johnsnowlabs.utils.testing import test_settings, nb_code_match


def flatten(text):
    return [item for sublist in text for item in sublist]


class JslPreprocessor(ExecutePreprocessor):
    def preprocess(self, nb: NotebookNode, resources=None, km=None):
        """
        See ExecutePreprocessor.preprocess for more infos
        """
        NotebookClient.__init__(self, nb, km)
        self.reset_execution_trackers()
        self._check_assign_resources(resources)

        # Inject custom nodes by prepending them
        for inject_node in test_settings.inject_header_nodes:
            nb["cells"].insert(0, inject_node)

        with self.setup_kernel():
            info_msg = self.wait_for_reply(self.kc.kernel_info())
            self.nb.metadata["language_info"] = info_msg["content"]["language_info"]
            for index, cell in enumerate(self.nb.cells):
                self.preprocess_cell(cell, resources, index)
        self.set_widgets_metadata()

        return self.nb, self.resources

    def preprocess_cell(self, cell, resources, index):
        """
        See ExecutePreprocessor.preprocess_cell for more infos
        See test_settings.py for all filter settings.
        Preprocess filters with the following rules :
        1. Replace line on Equality Match
        2. Replace line on Sub-String Match
        3. Drop line on Equality Match
        4. Drop line on Sub-String Match
        5. Drop on Regex Match
        """

        if cell["cell_type"] == "code":
            # Full Cell Code based Preprocessing
            cell_code = nb_code_match.replace_on_sub_string_match(
                cell["source"], test_settings.sub_string_replacement_cells
            )
            code_lines = cell_code.split("\n")
            new_lines = []
            # Single Line based Preprocessing - Apply rules to each line individually
            for code_line in code_lines:
                # Replacements for singe lines of code
                code_line = nb_code_match.replace_on_equal_string_match(
                    code_line,
                    test_settings.equal_string_replacement_lines,
                    log_type="line",
                )

                code_line = nb_code_match.replace_on_sub_string_match(
                    code_line,
                    test_settings.sub_string_replacement_lines,
                    log_type="line",
                )

                # Dropping Lines of code
                if nb_code_match.equals_sub_string(
                    code_line, test_settings.equal_string_drop_lines
                ):
                    print(f"[JSL-PREPROCESSOR]: DROPPING <<<{code_line}>>>")
                    continue

                # Sub-String Check
                if nb_code_match.contains_sub_string(
                    code_line, test_settings.sub_string_drop_lines
                ):
                    print(f"[JSL-PREPROCESSOR]: DROPPING <<<{code_line}>>>")
                    continue

                new_lines.append(code_line)

            new_lines = "\n".join(new_lines)

            # Clean anything that matches the regexes
            matches_to_clean = [
                re.findall(r, new_lines, re.DOTALL)
                for r in test_settings.regex_cell_content_drop_matcher
            ]
            for m in flatten(matches_to_clean):
                new_lines = new_lines.replace(m, "")

            cell["source"] = new_lines
            if test_settings.print_cell_before_executing:
                print(
                    f'[JSL-PREPROCESSOR]: EXECUTING CELL \n <<<{cell["source"]}>>> \n'
                )
            self._check_assign_resources(resources)
            cell = self.execute_cell(cell, index, store_history=True)
            return cell, self.resources

        return super().preprocess_cell(cell, resources, index)



import os
import time
from typing import Optional, List, Dict

from johnsnowlabs import settings
from johnsnowlabs.auto_install.jsl_home import get_install_suite_from_jsl_home
from johnsnowlabs.auto_install.softwares import Software
from johnsnowlabs.py_models.install_info import InstallSuite
from johnsnowlabs.utils.enums import JvmHardwareTarget


def authenticate_enviroment_HC(suite: InstallSuite):
    """Set Secret environ variables for Spark Context"""
    if suite.secrets.HC_LICENSE:
        os.environ["SPARK_NLP_LICENSE"] = suite.secrets.HC_LICENSE
    os.environ["AWS_ACCESS_KEY_ID"] = suite.secrets.AWS_ACCESS_KEY_ID
    os.environ["AWS_SECRET_ACCESS_KEY"] = suite.secrets.AWS_SECRET_ACCESS_KEY


def authenticate_enviroment_OCR(suite: InstallSuite):
    """Set Secret environ variables for Spark Context"""
    if suite.secrets.OCR_LICENSE:
        os.environ["SPARK_NLP_LICENSE"] = suite.secrets.OCR_LICENSE
    os.environ["AWS_ACCESS_KEY_ID"] = suite.secrets.AWS_ACCESS_KEY_ID
    os.environ["AWS_SECRET_ACCESS_KEY"] = suite.secrets.AWS_SECRET_ACCESS_KEY


def authenticate_enviroment_HC_and_OCR(suite: InstallSuite):
    """Set Secret environ variables for Spark Context"""
    authenticate_enviroment_HC(suite)
    authenticate_enviroment_OCR(suite)


def retry(fun, max_tries=10):
    for i in range(max_tries):
        try:
            time.sleep(0.3)
            fun()
            break
        except Exception:
            continue


def start(
    # -- JSL-Auth Flows --
    # Browser Auth
    browser_login: bool = False,
    # JWT Token Auth
    access_token: Optional[str] = None,
    # JSON file Auth
    json_license_path: Optional[str] = None,
    # AWS Auth
    aws_access_key: Optional[str] = None,
    aws_key_id: Optional[str] = None,
    # Manual License specification Auth
    enterprise_nlp_secret: Optional[str] = None,
    visual_secret: Optional[str] = None,
    hc_license: Optional[str] = None,
    ocr_license: Optional[str] = None,
    fin_license: Optional[str] = None,
    leg_license: Optional[str] = None,
    # License usage & Caching
    remote_license_number: int = 0,
    local_license_number: int = 0,
    store_in_jsl_home: bool = True,
    # -- Spark Session Configs --
    spark_conf: Optional[Dict[str, str]] = None,
    master_url: str = "local[*]",
    jar_paths: List[str] = None,
    # spark_nlp corresponds to enterprise NLP jar
    spark_nlp: bool = True,
    nlp: bool = True,
    visual: bool = False,
    hardware_target: str = JvmHardwareTarget.cpu.value,
    model_cache_folder: str = None,
    create_jsl_home_if_missing: bool = True,
) -> "pyspark.sql.SparkSession":
    from pyspark.sql import SparkSession

    already_launched = False
    if (
        "_instantiatedSession" in dir(SparkSession)
        and SparkSession._instantiatedSession is not None
    ):
        print("Spark Session already created, some configs may not take.")
        already_launched = True
        if settings.on_databricks:
            print(
                "Looks like you are on databricks which has a session pre-started. jsl.start() is not required. "
            )

    launched_products: List[str] = []
    hardware_target = JvmHardwareTarget.from_str(hardware_target)

    # Get all Local Jar Paths, downloads them if missing
    suite = get_install_suite_from_jsl_home(
        create_jsl_home_if_missing=create_jsl_home_if_missing,
        only_jars=True,
        jvm_hardware_target=hardware_target,
        force_browser=browser_login,
        browser_login=browser_login,
        access_token=access_token,
        local_license_number=local_license_number,
        remote_license_number=remote_license_number,
        secrets_file=json_license_path,
        hc_license=hc_license,
        hc_secret=enterprise_nlp_secret,
        ocr_secret=visual_secret,
        ocr_license=ocr_license,
        aws_access_key=aws_access_key,
        aws_key_id=aws_key_id,
        fin_license=fin_license,
        leg_license=leg_license,
        store_in_jsl_home=store_in_jsl_home,
    )

    # Collect all local Jar Paths we have access to for the SparkSession
    jars = []
    if (
        spark_nlp
        and Software.spark_nlp.check_installed(None)
        and suite.nlp.get_java_path()
    ):
        jars.append(suite.nlp.get_java_path())
        import sparknlp

        launched_products.append(
            f"{Software.spark_nlp.logo}{Software.spark_nlp.name}=={sparknlp.version()}"
        )

    if suite.secrets or not settings.license_required:
        if (
            suite.hc
            and nlp
            and Software.spark_hc.check_installed(None)
            and suite.hc.get_java_path()
        ):
            jars.append(suite.hc.get_java_path())
            if settings.license_required:
                authenticate_enviroment_HC(suite)
            import sparknlp_jsl

            launched_products.append(
                f"{Software.spark_hc.logo}{Software.spark_hc.name}=={sparknlp_jsl.version()}"
            )

        if (
            suite.ocr
            and visual
            and Software.spark_ocr.check_installed(None)
            and suite.ocr.get_java_path()
        ):
            jars.append(suite.ocr.get_java_path())
            if settings.license_required:
                authenticate_enviroment_OCR(suite)
            import sparkocr

            launched_products.append(
                f"{Software.spark_ocr.logo}{Software.spark_ocr.name}=={sparkocr.version()}"
            )
    import pyspark

    launched_products.append(
        f"running on {Software.spark.logo}{Software.pyspark.name}=={pyspark.version.__version__}"
    )

    builder = SparkSession.builder.appName(
        f'{settings.spark_session_name} with Jars for: {", ".join(launched_products)}'
    ).master(master_url)

    if jar_paths:
        # Add user specified Jars
        jars += jar_paths
    default_conf = {
        "spark.driver.memory": "16G",
        "spark.serializer": "org.apache.spark.serializer.KryoSerializer",
        "spark.kryoserializer.buffer.max": "2000M",
        "spark.driver.maxResultSize": "2000M",
        "spark.jars": ",".join(jars),
    }

    if suite.ocr and suite.ocr.get_java_path():
        # is_spark_version_env('32')
        default_conf["spark.sql.optimizer.expression.nestedPruning.enabled"] = "false"
        default_conf["spark.sql.optimizer.nestedSchemaPruning.enabled"] = "false"
        default_conf["spark.sql.legacy.allowUntypedScalaUDF"] = "true"
        default_conf["spark.sql.repl.eagerEval.enabled"] = "true"

    for k, v in default_conf.items():
        builder.config(str(k), str(v))

    if model_cache_folder:
        if not spark_conf:
            spark_conf = {}
        spark_conf["spark.jsl.settings.pretrained.cache_folder"] = model_cache_folder

    if spark_conf:
        for k, v in spark_conf.items():
            builder.config(str(k), str(v))
    spark = builder.getOrCreate()

    if suite.hc and exist_in_jvm(
        "com.johnsnowlabs.util.start.registerListenerAndStartRefresh"
    ):
        spark._jvm.com.johnsnowlabs.util.start.registerListenerAndStartRefresh()
    if suite.ocr and exist_in_jvm(
        "com.johnsnowlabs.util.OcrStart.registerListenerAndStartRefresh"
    ):
        retry(spark._jvm.com.johnsnowlabs.util.OcrStart.registerListenerAndStartRefresh)

    from colorama import Fore

    if not already_launched:
        print(
            f'👌 Launched {Fore.LIGHTGREEN_EX + hardware_target.value} optimized{Fore.RESET} session with with: {", ".join(launched_products)}'
        )

    return spark


def exist_in_jvm(java_class):
    # Check if a class path is present in JVM.
    # Only call this after Spark Session is started
    from pyspark.ml.util import _jvm
    from py4j.java_gateway import UserHelpAutoCompletion

    java_obj = _jvm()
    for name in java_class.split("."):
        # Bug in P4J. Even if ClassPath does not Exist, JVM response is proto.SUCCESS_PACKAGE
        # But it should give exception
        java_obj = getattr(java_obj, name)

    if UserHelpAutoCompletion.KEY in dir(java_obj):
        return False
    return True



from abc import ABCMeta

from johnsnowlabs.abstract_base.lib_resolver import (
    Py4JJslLibDependencyResolverABC,
    PyInstallTypes,
)
from johnsnowlabs.py_models.url_dependency import UrlDependency
from johnsnowlabs.utils.enums import (
    LatestCompatibleProductVersion,
    ProductName,
    SparkVersion,
    JvmHardwareTarget,
)


class NlpLibResolver(Py4JJslLibDependencyResolverABC, metaclass=ABCMeta):
    has_m1_jar = True
    has_cpu_jars = True
    has_py_install = True
    has_gpu_jars = True
    product_name = ProductName.nlp
    lib_version = LatestCompatibleProductVersion.spark_nlp.value
    compatible_spark_versions = [SparkVersion.spark3xx.value]

    compatible_spark_to_jar_map = {
        SparkVersion.spark3xx: {
            JvmHardwareTarget.gpu: UrlDependency(
                url="https://s3.amazonaws.com/auxdata.johnsnowlabs.com/public/jars/spark-nlp-gpu-assembly-{lib_version}.jar",
                dependency_type=JvmHardwareTarget.gpu,
                spark_version=SparkVersion.spark3xx,
                product_name=product_name,
                file_name=product_name.name,
                dependency_version=lib_version,
            ),
            JvmHardwareTarget.m1: UrlDependency(
                url="https://s3.amazonaws.com/auxdata.johnsnowlabs.com/public/jars/spark-nlp-m1-assembly-{lib_version}.jar",
                dependency_type=JvmHardwareTarget.m1,
                spark_version=SparkVersion.spark3xx,
                product_name=product_name,
                file_name=product_name.name,
                dependency_version=lib_version,
            ),
            JvmHardwareTarget.cpu: UrlDependency(
                url="https://s3.amazonaws.com/auxdata.johnsnowlabs.com/public/jars/spark-nlp-assembly-{lib_version}.jar",
                dependency_type=JvmHardwareTarget.cpu,
                spark_version=SparkVersion.spark3xx,
                product_name=product_name,
                file_name=product_name.name,
                dependency_version=lib_version,
            ),
        }
    }
    compatible_spark_to_py_map = {
        SparkVersion.spark3xx: {
            # TODO HARDCODE HASH!!! OR grap from enum or somwhere comfy. Maybe configs/settings file?
            PyInstallTypes.wheel: UrlDependency(
                url="https://files.pythonhosted.org/packages/65/19/c439d42f7afd75d6c9c20207db8ee0c95d7c82177b759303c7601120e91a/spark_nlp-{lib_version}-py2.py3-none-any.whl",
                dependency_type=PyInstallTypes.wheel,
                spark_version=SparkVersion.spark3xx,
                product_name=product_name,
                file_name=product_name.name,
                dependency_version=lib_version,
            ),
            PyInstallTypes.tar: UrlDependency(
                url="https://files.pythonhosted.org/packages/62/86/6b6c79f923db6ece28dd1c96d088fd2cc01ef7c748021ecb5fe73355635a/spark-nlp-{lib_version}.tar.gz",
                dependency_type=PyInstallTypes.tar,
                spark_version=SparkVersion.spark3xx,
                product_name=product_name,
                file_name=product_name.name,
                dependency_version=lib_version,
            ),
        }
    }
